/*
 * Copyright (c) 2002-2013, Mairie de Paris
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice
 *     and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright notice
 *     and the following disclaimer in the documentation and/or other materials
 *     provided with the distribution.
 *
 *  3. Neither the name of 'Mairie de Paris' nor 'Lutece' nor the names of its
 *     contributors may be used to endorse or promote products derived from
 *     this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * License 1.0
 */
package fr.paris.lutece.plugins.mvc;

import fr.paris.lutece.plugins.mvc.annotations.Action;
import fr.paris.lutece.plugins.mvc.annotations.Controller;
import fr.paris.lutece.plugins.mvc.annotations.View;
import fr.paris.lutece.portal.service.message.SiteMessageException;
import fr.paris.lutece.portal.service.plugin.Plugin;
import fr.paris.lutece.portal.service.template.AppTemplateService;
import fr.paris.lutece.portal.service.util.AppException;
import fr.paris.lutece.portal.service.util.AppLogService;
import fr.paris.lutece.portal.service.util.AppPropertiesService;
import fr.paris.lutece.portal.web.LocalVariables;
import fr.paris.lutece.portal.web.xpages.XPage;
import fr.paris.lutece.portal.web.xpages.XPageApplication;
import fr.paris.lutece.util.ErrorMessage;
import fr.paris.lutece.util.bean.BeanUtil;
import fr.paris.lutece.util.beanvalidation.BeanValidationUtil;
import fr.paris.lutece.util.html.HtmlTemplate;
import fr.paris.lutece.util.url.UrlItem;

import org.springframework.util.ReflectionUtils;

import java.io.IOException;

import java.lang.reflect.Method;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import javax.validation.ConstraintViolation;


/**
 * MVC XPage Application
 */
public abstract class MVCApplication implements XPageApplication
{
    private static final String MARK_ERRORS = "errors";
    private static final String URL_PORTAL = "Portal.jsp";
    private static final String PARAMETER_PAGE = "page";
    private static final String PARAMETER_VIEW = "view";
    private static final String PARAMETER_ACTION = "action";
    private List<ErrorMessage> _listErrors = new ArrayList<ErrorMessage>(  );
    private Controller _controller = getClass(  ).getAnnotation( Controller.class );

    /**
     * Returns the content of the page
     *
     * @param request The http request
     * @param nMode The current mode
     * @param plugin The plugin object
     * @return The XPage
     * @throws fr.paris.lutece.portal.service.message.SiteMessageException
     * Message displayed if an exception occurs
     */
    @Override
    public XPage getPage( HttpServletRequest request, int nMode, Plugin plugin )
        throws SiteMessageException
    {
        return processController( request );
    }

    ////////////////////////////////////////////////////////////////////////////
    // Controller

    /**
     * XPage controller
     * @param request The HTTP request
     * @return The XPage
     */
    private XPage processController( HttpServletRequest request )
    {
        Method[] methods = ReflectionUtils.getAllDeclaredMethods( getClass(  ) );

        try
        {
            // Process views
            Method m = findViewAnnotedMethod( methods, getView( request ) );

            if ( m != null )
            {
                return (XPage) m.invoke( this, request );
            }

            // Process actions
            m = findActionAnnotedMethod( methods, getAction( request ) );

            if ( m != null )
            {
                return (XPage) m.invoke( this, request );
            }

            // No view or action found so display the default view
            m = findDefaultViewMethod( methods );

            return (XPage) m.invoke( this, request );
        }
        catch ( Exception e )
        {
            throw new AppException( "MVC Error dispaching view and action ", e );
        }
    }

    /**
     * Get the view parameter
     * @param request The HTTP request
     * @return The view parameter or null if not found
     */
    protected String getView( HttpServletRequest request )
    {
        return request.getParameter( PARAMETER_VIEW );
    }

    /**
     * Get the action parameter
     * @param request The HTTP request
     * @return The action parameter or null if not found
     */
    protected String getAction( HttpServletRequest request )
    {
        return request.getParameter( PARAMETER_ACTION );
    }

    /**
     * Find the method that provide a given view
     * @param methods The methods list
     * @param strView The view
     * @return The method
     */
    private Method findViewAnnotedMethod( Method[] methods, String strView )
    {
        if ( strView != null )
        {
            for ( Method m : methods )
            {
                if ( m.isAnnotationPresent( View.class ) && strView.equals( m.getAnnotation( View.class ).value(  ) ) )
                {
                    return m;
                }
            }
        }

        return null;
    }

    /**
     * Find the method that provide a given action
     * @param methods The methods list
     * @param strAction The action
     * @return The method
     */
    private Method findActionAnnotedMethod( Method[] methods, String strAction )
    {
        if ( strAction != null )
        {
            for ( Method m : methods )
            {
                if ( m.isAnnotationPresent( Action.class ) &&
                        strAction.equals( m.getAnnotation( Action.class ).value(  ) ) )
                {
                    return m;
                }
            }
        }

        return null;
    }

    /**
     * Find the method that provide the default view
     * @param methods The methods list
     * @return The method
     */
    private Method findDefaultViewMethod( Method[] methods )
    {
        for ( Method m : methods )
        {
            if ( m.isAnnotationPresent( View.class ) && m.getAnnotation( View.class ).defaultView(  ) )
            {
                return m;
            }
        }

        return null;
    }

    /**
     * Returns the XPage name
     * @return The XPage name
     */
    protected String getXPageName(  )
    {
        return _controller.xpageName(  );
    }

    /**
     * Returns the default page title
     * @return The default page title
     */
    protected String getDefaultPageTitle(  )
    {
        return AppPropertiesService.getProperty( _controller.pageTitle(  ) );
    }

    /**
     * Returns the default page path
     * @return The default pagepath
     */
    protected String getDefaultPagePath(  )
    {
        return AppPropertiesService.getProperty( _controller.pagePath(  ) );
    }

    ////////////////////////////////////////////////////////////////////////////
    // XPage utils

    /**
     * Returns a new XPage object with default values
     * @return An XPage Object
     */
    protected XPage getXPage(  )
    {
        XPage page = new XPage(  );

        page.setTitle( getDefaultPageTitle(  ) );
        page.setPathLabel( getDefaultPagePath(  ) );

        return page;
    }

    /**
     * Returns a new XPage object with default values and the content filled
     * by a template
     * @param strTemplate The template
     * @return An XPage Object
     */
    protected XPage getXPage( String strTemplate )
    {
        XPage page = getXPage(  );

        HtmlTemplate t = AppTemplateService.getTemplate( strTemplate );
        page.setContent( t.getHtml(  ) );

        return page;
    }

    /**
     * Returns a new XPage object with default values and the content filled
     * by a template using a default model and for a given locale
     * @param strTemplate The template
     * @param locale The locale
     * @return An XPage Object
     */
    protected XPage getXPage( String strTemplate, Locale locale )
    {
        return getXPage( strTemplate, locale, getModel(  ) );
    }

    /**
     * Returns a new XPage object with default values and the content filled
     * by a template using a given model and for a given locale
     * @param strTemplate The template
     * @param locale The locale
     * @param model The model
     *
     * @return An XPage Object
     */
    protected XPage getXPage( String strTemplate, Locale locale, Map<String, Object> model )
    {
        XPage page = getXPage(  );

        HtmlTemplate t = AppTemplateService.getTemplate( strTemplate, locale, model );
        page.setContent( t.getHtml(  ) );

        return page;
    }

    /**
     * Get a model Object filled with default values
     * @return The model
     */
    protected Map<String, Object> getModel(  )
    {
        Map<String, Object> model = new HashMap<String, Object>(  );
        fillCommons( model );

        return model;
    }

    ////////////////////////////////////////////////////////////////////////////
    // Bean processing

    /**
     * Populate a bean using parameters in http request
     * @param bean bean to populate
     * @param request http request
     */
    protected void populate( Object bean, HttpServletRequest request )
    {
        BeanUtil.populate( bean, request );
    }

    /**
     * Validate a bean
     * @param <T> The bean class
     * @param bean The bean
     * @return true if validated otherwise false
     */
    protected <T> boolean validateBean( T bean )
    {
        _listErrors.clear(  );

        Set<ConstraintViolation<T>> errors = BeanValidationUtil.validate( bean );

        if ( errors.isEmpty(  ) )
        {
            return true;
        }

        for ( ConstraintViolation<T> constraint : errors )
        {
            MVCErrorMessage error = new MVCErrorMessage(  );
            error.setMessage( constraint.getMessage(  ) );
            _listErrors.add( error );
        }

        return false;
    }

    /**
     * Gets validation errors
     * @return The errors list
     */
    protected List<ErrorMessage> getValidationErrors(  )
    {
        return _listErrors;
    }

    /**
     * Fill the model with commons objects used in templates
     * @param model The model
     */
    protected void fillCommons( Map<String, Object> model )
    {
        model.put( MARK_ERRORS, _listErrors );
    }

    ////////////////////////////////////////////////////////////////////////////
    // Redirect utils

    /**
     * Redirect to requested page
     *
     * @param request the http request
     * @param strTarget the targeted page
     * @return the page requested
     */
    protected XPage redirect( HttpServletRequest request, String strTarget )
    {
        HttpServletResponse response = LocalVariables.getResponse(  );

        try
        {
            System.out.println( "Redirect :" + strTarget );
            response.sendRedirect( strTarget );
        }
        catch ( IOException e )
        {
            AppLogService.error( "Unable to redirect : " + strTarget + " : " + e.getMessage(  ), e );
        }

        return new XPage(  );
    }

    /**
     * Redirect to requested view
     *
     * @param request the http request
     * @param strView  the targeted view
     * @return the page requested
     */
    protected XPage redirectView( HttpServletRequest request, String strView )
    {
        UrlItem url = new UrlItem( URL_PORTAL );
        url.addParameter( PARAMETER_PAGE, getXPageName(  ) );
        url.addParameter( PARAMETER_VIEW, strView );

        return redirect( request, url.getUrl(  ) );
    }
}
